#define _GNU_SOURCE
#include <assert.h>
#include <err.h>
#include <fcntl.h>
#include <mqueue.h>
#include <openssl/rand.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>

#include "interface.h"

void spawn_server_instance() {
  pid_t pid = fork();
  if (pid == -1) {
    err(EXIT_FAILURE, "Failed to fork for the server process");
  } else if (pid == 0) {
    char *argv[] = {"/home/server/spawn_server", 0};
    if (execv(argv[0], &argv[0])) {
      err(EXIT_FAILURE, "Failed to spawn spawn_server process");
    }
  }
}

int64_t balance() {
  struct client_server_message msg = {.type = C2S_BALANCE};
  send(msg, "Failed to send balance request");

  struct server_client_message recv_msg;
  recv(recv_msg, "Failed to receive balance");
  assert(recv_msg.type == S2C_BALANCE);

  return recv_msg.message.balance;
}

void flip() {
  struct client_server_message msg = {.type = C2S_REQUEST_FLIP};
  send(msg, "Failed to request flip");

  struct server_client_message recv_msg;
  recv(recv_msg, "Failed to receive flip response");
  assert(recv_msg.type == S2C_FLIPPED);
}

struct server_client_message predict(enum coin_flip_result prediction,
                                     int64_t bet_amount) {
  struct client_server_message msg = {
      .type = C2S_FLIP_PREDICTION,
      .message.prediction = {.prediction = prediction, .bet = bet_amount}};
  send(msg, "Failed to send prediction");

  struct server_client_message recv_msg;
  recv(recv_msg, "Failed to receive prediction result");
  assert(recv_msg.type == S2C_PREDICTION_RESULT);
  return recv_msg;
}

bool predict_choice(char *choice) {
  char prediction;
  int64_t bet;
  if (sscanf(choice, "%c %ld\n", &prediction, &bet) != 2) {
    errx(EXIT_FAILURE, "Failed to parse prediction: %s\n", choice);
  }

  enum coin_flip_result p;
  switch (prediction) {
  case 'h':
    p = HEADS;
    break;
  case 't':
    p = TAILS;
    break;
  default:
    err(EXIT_FAILURE, "Unknown prediction: %c", prediction);
  }

  struct server_client_message m = predict(p, bet);
  return m.message.prediction_result.correct;
}

void buy_flag() {
  int64_t bal = balance();
  if (bal < FLAG_PRICE) {
    printf("Balance is too low! You need at least %ld ¤\n", FLAG_PRICE);
  } else {
    struct client_server_message m = {.type = C2S_BUY_FLAG};
    send(m, "Failed to send buy flag command");

    struct server_client_message r;
    recv(r, "Failed to receive flag");
    assert(r.type == S2C_FLAG);

    printf("Flag: %s\n", r.message.flag);
    exit(0);
  }
}

void print_options() {
  printf("\nOptions\n");
  printf("  b         \trequest your current balance\n");
  printf("  f         \tbuy the flag\n");
  printf("  r         \trequest a coin flip\n");
  printf("  h [amount]\tbet [amount] that the flip was heads\n");
  printf("  t [amount]\tbet [amount] that the flip was tails\n");
  printf("  e         \texit\n");
  printf("Choice: ");
}

int main() {
  spawn_server_instance();

  printf("                 .####*    *####.                 \n"
         "             ###  ##############  ###             \n"
         "           #- #####+##########:##### -#           \n"
         "        =# -## #### ########## #### ##- #-        \n"
         "       # +## ## ################## # ###= #       \n"
         "     ## ### #######            ######## ## ##     \n"
         "    ## # ########                #######  # ##    \n"
         "   ## #-#######                    ####### # ##   \n"
         "   # ### #####                      ######### #   \n"
         "  # ##########                      ######=### #  \n"
         "  # # # #####                        ##### # # #  \n"
         "  # #########   #####.       #####   ######### #  \n"
         "  # ## ######*  #######    #######  ###### ### #  \n"
         "  # #   ######    -#####  #####+    ###### ### #  \n"
         "  # ###########                    ########### #  \n"
         "  # ############                  ############ #  \n"
         "   # #  #########                #########  # #   \n"
         "    # ### ########              ######## ### #    \n"
         "    +# ###  ########          ########  ### #=    \n"
         "      # ####:#########      ########## ### #      \n"
         "       #= ############################## =#       \n"
         "         #  ### +## ##########=-#* ###  #         \n"
         "           ##  ### #####.#*#  # ###  ##           \n"
         "              ##:  ############  :##              \n"
         "                  *############*                  \n"
         "                                                  \n");

  printf(
      "The aliens wants to play a game with you, a betting game. They\n"
      "try to explain it but the language barrier makes it very difficult to\n"
      "understand what they mean. Luckily for you, your AI magic translator\n"
      "already understands the alien speech!\n\n"
      "'Predict the result of the coin flip. If you predicted the result\n"
      "correctly, then you will win 2x your bet! If your balance reaches\n"
      "\t%ld ¤\n"
      "then you can throw away all your ¤ and buy the flag!'\n\n",
      FLAG_PRICE);

  printf("(please note: the intended solution does not do any pwn)\n");

  setup_shared_memory_client();
  setup_message_queues_client();

  while (1) {
    print_options();

    char *lineptr = NULL;
    size_t n = 0;
    ssize_t length = 0;
    if (length = getline(&lineptr, &n, stdin), length == -1) {
      exit(1);
    }

    if (!length) {
      free(lineptr);
      continue;
    }

    printf("\n");

    int64_t bal;
    switch (lineptr[0]) {

    case 'b':
      bal = balance();
      printf("Balance: %ld ¤\n", bal);
      break;

    case 'f':
      buy_flag();
      break;

    case 'r':
      flip();
      printf("Flipped!\n");
      break;

    case 'h':
    case 't':
      if (predict_choice(lineptr)) {
        printf("Good prediction! Doubled your bet!\n");
      } else {
        printf("WRONG! Lost your bet\n");
      }
      break;

    case 'e':
      exit(0);
      break;

    default:
      break;
    }

    free(lineptr);
  }

  printf("Success!\n");
  return 0;
}
