#pragma once
#include <err.h>
#include <errno.h>
#include <mqueue.h>
#include <sched.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>

static const size_t PAGE_SIZE = 4096;
static const char *const SHM_NAME = "/client_server_shm";
static const char *const MQ_NAME_SERVER_CLIENT = "/server_client";
static const char *const MQ_NAME_CLIENT_SERVER = "/client_server";
static const size_t SHM_SIZE = 2 * PAGE_SIZE;
static const int64_t FLAG_PRICE = 1000000000000000000;
static void *const SHM_ADDR = (void *)0x7deadbeef000;

////////////////////////////////////
////////// MESSAGE QUEUES //////////
////////////////////////////////////
// The following definitions are used to setup the message queues
// between the client and the server, so they can communicate with
// each other.

static mqd_t send_queue = {0};
static mqd_t recv_queue = {0};

#define send(msg, FMT)                                                         \
  if (mq_send(send_queue, (char *)&msg, sizeof(msg), 0) == -1) {               \
    err(EXIT_FAILURE, FMT);                                                    \
  }

#define recv(msg, FMT)                                                         \
  if (mq_receive(recv_queue, (char *)&msg, sizeof(msg), NULL) == -1) {         \
    err(EXIT_FAILURE, FMT);                                                    \
  }

enum coin_flip_result { HEADS, TAILS };

enum s2c_message_type {
  S2C_BALANCE,
  S2C_FLIPPED,
  S2C_PREDICTION_RESULT,
  S2C_FLAG
};

struct server_client_message {
  enum s2c_message_type type;
  union {
    struct {
      bool correct;
    } prediction_result;
    int64_t balance;
    char *flag;
  } message;
};

enum c2s_message_type {
  C2S_BALANCE,
  C2S_REQUEST_FLIP,
  C2S_FLIP_PREDICTION,
  C2S_BUY_FLAG
};

struct client_server_message {
  enum c2s_message_type type;
  union {
    struct {
      enum coin_flip_result prediction;
      int64_t bet;
    } prediction;
  } message;
};

// Setup the message queues on the server side, or create the
// message queues if they don't exist yet.
//
// The message queues show up here
//
// client@machine:~$ ls -lah /dev/mqueue/
// -r------w- 1 server server  80 May  1 19:53 client_server
// --w----r-- 1 server server  80 May  1 19:53 server_client
//
void setup_message_queues_server() {
  struct mq_attr attrs_s2c = {
      .mq_maxmsg = 5, .mq_msgsize = sizeof(struct server_client_message)};

  mode_t old_umask = umask(0);

  send_queue = mq_open(MQ_NAME_SERVER_CLIENT, O_CREAT | O_WRONLY,
                       S_IWUSR | S_IROTH, &attrs_s2c);
  if (send_queue == -1) {
    err(EXIT_FAILURE, "Failed to create and open message queue %s",
        MQ_NAME_SERVER_CLIENT);
  }

  struct mq_attr attrs_c2s = {
      .mq_maxmsg = 5, .mq_msgsize = sizeof(struct client_server_message)};
  recv_queue = mq_open(MQ_NAME_CLIENT_SERVER, O_CREAT | O_RDONLY,
                       S_IRUSR | S_IWOTH, &attrs_c2s);
  if (recv_queue == -1) {
    err(EXIT_FAILURE, "Failed to create and open message queue %s",
        MQ_NAME_SERVER_CLIENT);
  }

  umask(old_umask);
}

// Connect to the message queues on the client side
void setup_message_queues_client() {
  recv_queue = mq_open(MQ_NAME_SERVER_CLIENT, O_RDONLY);
  if (recv_queue == -1) {
    err(EXIT_FAILURE, "Failed to open server->client message queue");
  }

  send_queue = mq_open(MQ_NAME_CLIENT_SERVER, O_WRONLY);
  if (send_queue == -1) {
    err(EXIT_FAILURE, "Failed to open client->server message queue");
  }
}

////////////////////////////////////
////////// SHARED MEMORY ///////////
////////////////////////////////////
// Setup some shared memory pages that are writable by the server,
// and read only by the client.

volatile void *setup_shared_memory_server() {
  int fd;
  do {
    fd = shm_open(SHM_NAME, O_RDWR | O_CREAT | O_EXCL | O_TRUNC,
                  S_IRUSR | S_IWUSR | S_IROTH);

    if (fd == -1) {
      if (errno == EEXIST) {
        if (shm_unlink(SHM_NAME) == -1) {
          err(EXIT_FAILURE, "Failed to unlink shared memory %s", SHM_NAME);
        }
      }
    }
  } while (fd == -1);

  if (ftruncate(fd, SHM_SIZE) == -1) {
    err(EXIT_FAILURE, "Failed to resize shared memory file");
  }

  char *shared_mem = mmap(SHM_ADDR, SHM_SIZE, PROT_READ | PROT_WRITE,
                          MAP_SHARED | MAP_POPULATE | MAP_FIXED, fd, 0);
  if (shared_mem == MAP_FAILED) {
    err(EXIT_FAILURE, "Failed to mmap shared memory");
  }

  memcpy(&shared_mem[0], "\tTAILS!\0", 8);
  memcpy(&shared_mem[4096], "\tHEADS!\0", 8);

  return shared_mem;
}

volatile void *setup_shared_memory_client() {
  int fd;
  do {
    fd = shm_open(SHM_NAME, O_RDONLY, 0);
    if (fd == -1) {
      if (errno != ENOENT) {
        err(EXIT_FAILURE, "Failed to open shared memory file %s", SHM_NAME);
      } else {
        sched_yield();
      }
    }
  } while (fd == -1);

  void *shared_mem = mmap(SHM_ADDR, SHM_SIZE, PROT_READ,
                          MAP_SHARED | MAP_POPULATE | MAP_FIXED, fd, 0);
  if (shared_mem == MAP_FAILED) {
    err(EXIT_FAILURE, "Failed to mmap shared memory");
  }

  return shared_mem;
}
